# -*- coding: utf-8 -*-
# Generated by Django 1.11.6 on 2017-11-01 15:58
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('wagtail_personalisation', '0013_add_dynamic_static_to_segment'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='segment',
            name='sessions',
        ),
        migrations.AddField(
            model_name='segment',
            name='static_users',
            field=models.ManyToManyField(to=settings.AUTH_USER_MODEL),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2016-12-05 16:23
from __future__ import unicode_literals

import django.db.models.deletion
import modelcluster.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wagtail_personalisation', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='QueryRule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('query_parameter', models.TextField(verbose_name='The query parameter to search for')),
                ('query_value', models.TextField(verbose_name='The value of the parameter to match')),
                ('segment', modelcluster.fields.ParentalKey(on_delete=django.db.models.deletion.CASCADE, related_name='wagtail_personalisation_queryrule_related', related_query_name='wagtail_personalisation_queryrules', to='wagtail_personalisation.Segment')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AlterField(
            model_name='visitcountrule',
            name='operator',
            field=models.CharField(choices=[('more_than', 'More than'), ('less_than', 'Less than'), ('equal_to', 'Equal to')], default='more_than', max_length=20),
        ),
    ]



# Generated by Django 2.0.7 on 2018-07-04 15:26

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('wagtail_personalisation', '0020_rules_delete_relatedqueryname'),
    ]

    operations = [
        migrations.AlterField(
            model_name='personalisablepagemetadata',
            name='segment',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, related_name='page_metadata', to='wagtail_personalisation.Segment'),
        ),
    ]



# Generated by Django 2.0.7 on 2018-07-05 13:25

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wagtail_personalisation', '0021_personalisablepagemetadata_segment_set_on_delete_protect'),
    ]

    operations = [
        migrations.AlterField(
            model_name='personalisablepagemetadata',
            name='canonical_page',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, related_name='personalisable_canonical_metadata', to='wagtailcore.Page'),
        ),
    ]



# Generated by Django 2.0.5 on 2018-05-26 14:25

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wagtail_personalisation', '0018_segment_excluded_users'),
    ]

    operations = [
        migrations.AlterField(
            model_name='personalisablepagemetadata',
            name='segment',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='page_metadata', to='wagtail_personalisation.Segment'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.3 on 2016-12-05 07:48
from __future__ import unicode_literals

import django.db.models.deletion
import modelcluster.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('wagtailcore', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='PersonalisablePage',
            fields=[
                ('page_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='wagtailcore.Page')),
                ('is_segmented', models.BooleanField(default=False)),
                ('canonical_page', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='variants', to='wagtail_personalisation.PersonalisablePage')),
            ],
            options={
                'abstract': False,
            },
            bases=('wagtailcore.page',),
        ),
        migrations.CreateModel(
            name='ReferralRule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('regex_string', models.TextField(verbose_name='Regex string to match the referer with')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Segment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('create_date', models.DateTimeField(auto_now_add=True)),
                ('edit_date', models.DateTimeField(auto_now=True)),
                ('enable_date', models.DateTimeField(editable=False, null=True)),
                ('disable_date', models.DateTimeField(editable=False, null=True)),
                ('visit_count', models.PositiveIntegerField(default=0, editable=False)),
                ('status', models.CharField(choices=[('enabled', 'Enabled'), ('disabled', 'Disabled')], default='enabled', max_length=20)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='TimeRule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('start_time', models.TimeField(verbose_name='Starting time')),
                ('end_time', models.TimeField(verbose_name='Ending time')),
                ('segment', modelcluster.fields.ParentalKey(on_delete=django.db.models.deletion.CASCADE, related_name='wagtail_personalisation_timerule_related', related_query_name='wagtail_personalisation_timerules', to='wagtail_personalisation.Segment')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='VisitCountRule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('operator', models.CharField(choices=[('more_than', 'More than'), ('less_than', 'Less than'), ('equal_to', 'Equal to')], default='ht', max_length=20)),
                ('count', models.PositiveSmallIntegerField(default=0, null=True)),
                ('counted_page', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='wagtailcore.Page')),
                ('segment', modelcluster.fields.ParentalKey(on_delete=django.db.models.deletion.CASCADE, related_name='wagtail_personalisation_visitcountrule_related', related_query_name='wagtail_personalisation_visitcountrules', to='wagtail_personalisation.Segment')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='referralrule',
            name='segment',
            field=modelcluster.fields.ParentalKey(on_delete=django.db.models.deletion.CASCADE, related_name='wagtail_personalisation_referralrule_related', related_query_name='wagtail_personalisation_referralrules', to='wagtail_personalisation.Segment'),
        ),
        migrations.AddField(
            model_name='personalisablepage',
            name='segment',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='segments', to='wagtail_personalisation.Segment'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-04-20 15:47
from __future__ import unicode_literals

import django.db.models.deletion
import modelcluster.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wagtail_personalisation', '0007_dayrule'),
    ]

    operations = [
        migrations.CreateModel(
            name='DeviceRule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('mobile', models.BooleanField(default=False, verbose_name='Mobile phone')),
                ('tablet', models.BooleanField(default=False, verbose_name='Tablet')),
                ('desktop', models.BooleanField(default=False, verbose_name='Desktop')),
                ('segment', modelcluster.fields.ParentalKey(on_delete=django.db.models.deletion.CASCADE, related_name='wagtail_personalisation_devicerule_related', related_query_name='wagtail_personalisation_devicerules', to='wagtail_personalisation.Segment')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-05-31 04:28
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('wagtail_personalisation', '0008_devicerule'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='personalisablepage',
            name='canonical_page',
        ),
        migrations.RemoveField(
            model_name='personalisablepage',
            name='page_ptr',
        ),
        migrations.RemoveField(
            model_name='personalisablepage',
            name='segment',
        ),
        migrations.DeleteModel(
            name='PersonalisablePage',
        ),
    ]



import time

from django.conf import settings
from django.db.models import F
from django.template.base import FilterExpression, kwarg_re
from django.utils import timezone
from django.utils.module_loading import import_string


def impersonate_other_page(page, other_page):
    """Function to change the page metadata so the user gets to see the
    non-personalized path and page.

    :param page: The page to be impersonated
    :type page: wagtail_personalisation.models.PersonalisablePage
    :param other_page: The page it should impersonate
    :type other_page: wagtail_personalisation.models.PersonalisablePage

    """
    page.path = other_page.path
    page.depth = other_page.depth
    page.url_path = other_page.url_path
    page.title = other_page.title


def create_segment_dictionary(segment):
    """Creates a dictionary with all the required segment information.

    :param segment: Segment object
    :type segment: wagtail_personalisation.models.Segment
    :return: Dictionary with name, id, timestamp and persistent state.
    :rtype: dict

    """
    return {
        "encoded_name": segment.encoded_name(),
        "id": segment.pk,
        "timestamp": int(time.time()),
        "persistent": segment.persistent
    }


def count_active_days(enable_date, disable_date):
    """Return the number of days the segment has been active.

    :param enable_date: The date the segment was enabled
    :type enable_date: timezone.datetime
    :param disable_date: The date the segment was disabled
    :type disable_date: timezone.datetime
    :returns: The amount of days a segment is/has been active
    :rtype: int

    """
    if enable_date is not None:
        if disable_date is None or disable_date <= enable_date:
            # There is no disable date, or it is not relevant.
            delta = timezone.now() - enable_date
            return delta.days
        if disable_date > enable_date:
            # There is a disable date and it is relevant.
            delta = disable_date - enable_date
            return delta.days

    return 0


def parse_tag(token, parser):
    """Parses template tag for name, arguments and keyword arguments.

    :param token: Template token containing all the tag contents
    :type token: django.template.base.Token
    :param parser: Template parser
    :type parser: django.template.base.Parser
    :return: Tuple with tag name, arguments and keyword arguments
    :rtype: tuple

    """
    # Split the tag content into words, respecting quoted strings.
    bits = token.split_contents()

    # Pull out the tag name.
    tag_name = bits.pop(0)

    # Parse the rest of the args, and build FilterExpressions from them so that
    # we can evaluate them later.
    args = []
    kwargs = {}
    for bit in bits:
        # Is this a kwarg or an arg?
        match = kwarg_re.match(bit)
        kwarg_format = match and match.group(1)
        if kwarg_format:
            key, value = match.groups()
            kwargs[key] = FilterExpression(value, parser)
        else:
            args.append(FilterExpression(bit, parser))

    return (tag_name, args, kwargs)


def exclude_variants(pages):
    """Checks if page is not a variant

    :param pages: Set of pages to check
    :type pages: QuerySet
    :return: Queryset of pages that aren't variants
    :rtype: QuerySet
    """
    from wagtail_personalisation.models import PersonalisablePageMetadata
    excluded_variant_pages = PersonalisablePageMetadata.objects.exclude(
        canonical_page_id=F('variant_id')
    ).values_list('variant_id')
    return pages.exclude(pk__in=excluded_variant_pages)


def can_delete_pages(pages, user):
    for variant in pages:
        if not variant.permissions_for_user(user).can_delete():
            return False
    return True


def get_client_ip(request):
    try:
        func = import_string(settings.WAGTAIL_PERSONALISATION_IP_FUNCTION)
    except AttributeError:
        pass
    else:
        return func(request)
    try:
        x_forwarded_for = request.META['HTTP_X_FORWARDED_FOR']
        return x_forwarded_for.split(',')[-1].strip()
    except KeyError:
        return request.META['REMOTE_ADDR']



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-05-31 11:01
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wagtail_personalisation', '0009_auto_20170531_0428'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='dayrule',
            options={'verbose_name': 'Day Rule'},
        ),
        migrations.AlterModelOptions(
            name='devicerule',
            options={'verbose_name': 'Device Rule'},
        ),
        migrations.AlterModelOptions(
            name='queryrule',
            options={'verbose_name': 'Query Rule'},
        ),
        migrations.AlterModelOptions(
            name='referralrule',
            options={'verbose_name': 'Referral Rule'},
        ),
        migrations.AlterModelOptions(
            name='timerule',
            options={'verbose_name': 'Time Rule'},
        ),
        migrations.AlterModelOptions(
            name='userisloggedinrule',
            options={'verbose_name': 'Logged in Rule'},
        ),
        migrations.AlterModelOptions(
            name='visitcountrule',
            options={'verbose_name': 'Visit count Rule'},
        ),
        migrations.AlterField(
            model_name='referralrule',
            name='regex_string',
            field=models.TextField(verbose_name='Regular expression to match the referrer'),
        ),
    ]



# Generated by Django 2.0.5 on 2018-05-30 18:51

import django.db.models.deletion
import modelcluster.fields
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('wagtail_personalisation', '0019_auto_20180526_1425'),
    ]

    operations = [
        migrations.AlterField(
            model_name='dayrule',
            name='segment',
            field=modelcluster.fields.ParentalKey(on_delete=django.db.models.deletion.CASCADE, related_name='wagtail_personalisation_dayrules', to='wagtail_personalisation.Segment'),
        ),
        migrations.AlterField(
            model_name='devicerule',
            name='segment',
            field=modelcluster.fields.ParentalKey(on_delete=django.db.models.deletion.CASCADE, related_name='wagtail_personalisation_devicerules', to='wagtail_personalisation.Segment'),
        ),
        migrations.AlterField(
            model_name='queryrule',
            name='segment',
            field=modelcluster.fields.ParentalKey(on_delete=django.db.models.deletion.CASCADE, related_name='wagtail_personalisation_queryrules', to='wagtail_personalisation.Segment'),
        ),
        migrations.AlterField(
            model_name='referralrule',
            name='segment',
            field=modelcluster.fields.ParentalKey(on_delete=django.db.models.deletion.CASCADE, related_name='wagtail_personalisation_referralrules', to='wagtail_personalisation.Segment'),
        ),
        migrations.AlterField(
            model_name='timerule',
            name='segment',
            field=modelcluster.fields.ParentalKey(on_delete=django.db.models.deletion.CASCADE, related_name='wagtail_personalisation_timerules', to='wagtail_personalisation.Segment'),
        ),
        migrations.AlterField(
            model_name='userisloggedinrule',
            name='segment',
            field=modelcluster.fields.ParentalKey(on_delete=django.db.models.deletion.CASCADE, related_name='wagtail_personalisation_userisloggedinrules', to='wagtail_personalisation.Segment'),
        ),
        migrations.AlterField(
            model_name='visitcountrule',
            name='segment',
            field=modelcluster.fields.ParentalKey(on_delete=django.db.models.deletion.CASCADE, related_name='wagtail_personalisation_visitcountrules', to='wagtail_personalisation.Segment'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.9 on 2018-02-09 08:28
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('wagtail_personalisation', '0017_segment_randomisation_percent'),
    ]

    operations = [
        migrations.AddField(
            model_name='segment',
            name='excluded_users',
            field=models.ManyToManyField(help_text='Users that matched the rules but were excluded from the segment for some reason e.g. randomisation', related_name='excluded_segments', to=settings.AUTH_USER_MODEL),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-06-01 11:48
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('wagtail_personalisation', '0011_personalisablepagemetadata'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='personalisablepagemetadata',
            name='is_segmented',
        ),
    ]



from __future__ import absolute_import, unicode_literals

from django.contrib import admin

from wagtail_personalisation import models, rules


class UserIsLoggedInRuleAdminInline(admin.TabularInline):
    """Inline the UserIsLoggedIn Rule into the
    administration interface for segments.

    """
    model = rules.UserIsLoggedInRule


class TimeRuleAdminInline(admin.TabularInline):
    """Inline the Time Rule into the
    administration interface for segments.

    """
    model = rules.TimeRule


class ReferralRuleAdminInline(admin.TabularInline):
    """Inline the Referral Rule into the
    administration interface for segments.

    """
    model = rules.ReferralRule


class VisitCountRuleAdminInline(admin.TabularInline):
    """Inline the Visit Count Rule into the
    administration interface for segments.

    """
    model = rules.VisitCountRule


class SegmentAdmin(admin.ModelAdmin):
    """Add the inline models to the Segment admin interface."""
    inlines = (UserIsLoggedInRuleAdminInline, TimeRuleAdminInline,
               ReferralRuleAdminInline, VisitCountRuleAdminInline)


admin.site.register(models.Segment, SegmentAdmin)



# Generated by Django 2.0.5 on 2018-07-19 09:57

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('wagtail_personalisation', '0022_personalisablepagemetadata_canonical_protect'),
    ]

    operations = [
        migrations.AlterField(
            model_name='personalisablepagemetadata',
            name='variant',
            field=models.OneToOneField(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='_personalisable_page_metadata', to='wagtailcore.Page'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.9 on 2018-01-25 09:18
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wagtail_personalisation', '0015_static_users'),
    ]

    operations = [
        migrations.AddField(
            model_name='segment',
            name='matched_count_updated_at',
            field=models.DateTimeField(editable=False, null=True),
        ),
        migrations.AddField(
            model_name='segment',
            name='matched_users_count',
            field=models.PositiveIntegerField(default=0, editable=False),
        ),
    ]



from django import template
from django.template import TemplateSyntaxError
from django.utils.safestring import mark_safe

from wagtail_personalisation.adapters import get_segment_adapter
from wagtail_personalisation.models import Segment
from wagtail_personalisation.utils import parse_tag

register = template.Library()


def do_segment(parser, token):
    """Block that only shows content if user is in chosen segment."""
    # Parse the tag
    tag_name, _, kwargs = parse_tag(token, parser)

    # If no segment is provided this block will raise an error
    if set(kwargs.keys()) != {'name'}:
        usage = '{% segment name="segmentname" %} ... {% endsegment %}'
        raise TemplateSyntaxError("Usage: %s" % usage)

    nodelist = parser.parse(('endsegment',))
    parser.delete_first_token()

    return SegmentNode(nodelist, name=kwargs['name'])


register.tag('segment', do_segment)


class SegmentNode(template.Node):
    """Node that only returns contents if user is in the segment.

    This node checks if the chosen segment exists and if the
    user has been segmented in the chosen segment.
    If not it will return nothing

    """
    def __init__(self, nodelist, name):
        self.nodelist = nodelist
        self.name = name

    def render(self, context):
        # Check if segment exists
        name = self.name.resolve(context)
        segment = Segment.objects.enabled().filter(name=name).first()
        if not segment:
            return ""

        # Check if user has segment
        adapter = get_segment_adapter(context['request'])
        user_segment = adapter.get_segment_by_id(segment_id=segment.pk)
        if not user_segment:
            return ''

        content = self.nodelist.render(context)
        content = mark_safe(content)
        return content



from __future__ import absolute_import, unicode_literals

from datetime import datetime
from importlib import import_module

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.contrib.staticfiles.templatetags.staticfiles import static
from django.test.client import RequestFactory
from django.utils.lru_cache import lru_cache
from django.utils.translation import ugettext_lazy as _
from wagtail.admin.forms import WagtailAdminModelForm

SessionStore = import_module(settings.SESSION_ENGINE).SessionStore


@lru_cache(maxsize=1000)
def user_from_data(user_id):
    User = get_user_model()
    try:
        return User.objects.get(id=user_id)
    except User.DoesNotExist:
        return AnonymousUser()


class SegmentAdminForm(WagtailAdminModelForm):

    def count_matching_users(self, rules, match_any):
        """ Calculates how many users match the given static rules
        """
        count = 0

        static_rules = [rule for rule in rules if rule.static]

        if not static_rules:
            return count

        User = get_user_model()
        users = User.objects.filter(is_active=True, is_staff=False)

        for user in users.iterator():
            if match_any:
                if any(rule.test_user(None, user) for rule in static_rules):
                    count += 1
            elif all(rule.test_user(None, user) for rule in static_rules):
                count += 1

        return count

    def clean(self):
        cleaned_data = super(SegmentAdminForm, self).clean()
        Segment = self._meta.model

        rules = [
            form.instance for formset in self.formsets.values()
            for form in formset
            if form not in formset.deleted_forms
        ]
        consistent = rules and Segment.all_static(rules)

        if cleaned_data.get('type') == Segment.TYPE_STATIC and not cleaned_data.get('count') and not consistent:
            self.add_error('count', _('Static segments with non-static compatible rules must include a count.'))

        if self.instance.id and self.instance.is_static:
            if self.has_changed():
                self.add_error_to_fields(self, excluded=['name', 'enabled'])

            for formset in self.formsets.values():
                if formset.has_changed():
                    for form in formset:
                        if form not in formset.deleted_forms:
                            self.add_error_to_fields(form)

        return cleaned_data

    def add_error_to_fields(self, form, excluded=list()):
        for field in form.changed_data:
            if field not in excluded:
                form.add_error(field, _('Cannot update a static segment'))

    def save(self, *args, **kwargs):
        is_new = not self.instance.id

        if not self.instance.is_static:
            self.instance.count = 0

        if is_new and self.instance.is_static and not self.instance.all_rules_static:
            rules = [
                form.instance for formset in self.formsets.values()
                for form in formset
                if form not in formset.deleted_forms
            ]
            self.instance.matched_users_count = self.count_matching_users(
                rules, self.instance.match_any)
            self.instance.matched_count_updated_at = datetime.now()

        instance = super(SegmentAdminForm, self).save(*args, **kwargs)

        if is_new and instance.is_static and instance.all_rules_static:
            from .adapters import get_segment_adapter

            request = RequestFactory().get('/')
            request.session = SessionStore()
            adapter = get_segment_adapter(request)

            users_to_add = []
            users_to_exclude = []

            User = get_user_model()
            users = User.objects.filter(is_active=True, is_staff=False)

            matched_count = 0
            for user in users.iterator():
                request.user = user
                passes = adapter._test_rules(instance.get_rules(), request, instance.match_any)
                if passes:
                    matched_count += 1
                    if instance.count == 0 or len(users_to_add) < instance.count:
                        if instance.randomise_into_segment():
                            users_to_add.append(user)
                        else:
                            users_to_exclude.append(user)

            instance.matched_users_count = matched_count
            instance.matched_count_updated_at = datetime.now()
            instance.static_users.add(*users_to_add)
            instance.excluded_users.add(*users_to_exclude)

        return instance

    @property
    def media(self):
        media = super(SegmentAdminForm, self).media
        media.add_js(
            [static('js/segment_form_control.js')]
        )
        return media



from __future__ import absolute_import, unicode_literals

from django.utils.translation import ugettext_lazy as _
from wagtail.core import blocks

from wagtail_personalisation.adapters import get_segment_adapter
from wagtail_personalisation.models import Segment


def list_segment_choices():
    yield -1, ("Show to everyone")
    for pk, name in Segment.objects.values_list('pk', 'name'):
        yield pk, name


class PersonalisedStructBlock(blocks.StructBlock):
    """Struct block that allows personalisation per block."""

    segment = blocks.ChoiceBlock(
        choices=list_segment_choices,
        required=False, label=_("Personalisation segment"),
        help_text=_("Only show this content block for users in this segment"))

    def render(self, value, context=None):
        """Only render this content block for users in this segment.

        :param value: The value from the block
        :type value: dict
        :param context: The context containing the request
        :type context: dict
        :returns: The provided block if matched, otherwise an empty string
        :rtype: blocks.StructBlock or empty str

        """
        request = context['request']
        adapter = get_segment_adapter(request)
        user_segments = adapter.get_segments()

        try:
            segment_id = int(value['segment'])
        except (ValueError, TypeError):
            return ''

        if segment_id > 0:
            for segment in user_segments:
                if segment.id == segment_id:
                    return super(PersonalisedStructBlock, self).render(
                        value, context)

        if segment_id == -1:
            return super(PersonalisedStructBlock, self).render(
                value, context)

        return ''



from django.db.models.signals import pre_save
from django.utils import timezone

from wagtail_personalisation.models import Segment


def check_status_change(sender, instance, *args, **kwargs):
    """Check if the status has changed. Alter dates accordingly."""
    try:
        original_status = sender.objects.get(pk=instance.id).status
    except sender.DoesNotExist:
        original_status = ""

    if original_status != instance.status:
        if instance.status == instance.STATUS_ENABLED:
            instance.enable_date = timezone.now()
            instance.visit_count = 0
            return instance
        if instance.status == instance.STATUS_DISABLED:
            instance.disable_date = timezone.now()


def register():
    pre_save.connect(check_status_change, sender=Segment)



from __future__ import absolute_import, unicode_literals

from django.conf.urls import url

from wagtail_personalisation import views

app_name = 'segment'

urlpatterns = [
    url(r'^segment/(?P<segment_id>[0-9]+)/toggle/$',
        views.toggle, name='toggle'),
    url(r'^(?P<page_id>[0-9]+)/copy/(?P<segment_id>[0-9]+)$',
        views.copy_page_view, name='copy_page'),
    url(r'^segment/toggle_segment_view/$',
        views.toggle_segment_view, name='toggle_segment_view'),
    url(r'^segment/users/(?P<segment_id>[0-9]+)$',
        views.segment_user_data, name='segment_user_data'),
]



from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class WagtailPersonalisationConfig(AppConfig):
    label = 'wagtail_personalisation'
    name = 'wagtail_personalisation'
    verbose_name = _('Wagtail Personalisation')

    def ready(self):
        from wagtail_personalisation import receivers

        receivers.register()



# -*- coding: utf-8 -*-
# Generated by Django 1.10.3 on 2016-12-06 10:05
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wagtail_personalisation', '0002_auto_20161205_1623'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='queryrule',
            name='query_parameter',
        ),
        migrations.RemoveField(
            model_name='queryrule',
            name='query_value',
        ),
        migrations.AddField(
            model_name='queryrule',
            name='parameter',
            field=models.SlugField(default='test', max_length=20, verbose_name='The query parameter to search for'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='queryrule',
            name='value',
            field=models.SlugField(default='test', max_length=20, verbose_name='The value of the parameter to match'),
            preserve_default=False,
        ),
    ]



from __future__ import absolute_import, unicode_literals

import logging

from django.conf.urls import include, url
from django.db import transaction
from django.http import Http404
from django.shortcuts import redirect, render
from django.template.defaultfilters import pluralize
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _
from wagtail.admin import messages
from wagtail.admin.site_summary import PagesSummaryItem, SummaryItem
from wagtail.admin.views.pages import get_valid_next_url_from_request
from wagtail.admin.widgets import Button, ButtonWithDropdownFromHook
from wagtail.core import hooks
from wagtail.core.models import Page

from wagtail_personalisation import admin_urls, models, utils
from wagtail_personalisation.adapters import get_segment_adapter

logger = logging.getLogger(__name__)


@hooks.register('register_admin_urls')
def register_admin_urls():
    """Adds the administration urls for the personalisation apps."""
    return [
        url(r'^personalisation/', include(
            admin_urls, namespace='wagtail_personalisation')),
    ]


@hooks.register('before_serve_page')
def set_visit_count(page, request, serve_args, serve_kwargs):
    """Tests the provided rules to see if the request still belongs
    to a segment.

    :param page: The page being served
    :type page: wagtail.core.models.Page
    :param request: The http request
    :type request: django.http.HttpRequest

    """
    adapter = get_segment_adapter(request)
    adapter.add_page_visit(page)


@hooks.register('before_serve_page')
def segment_user(page, request, serve_args, serve_kwargs):
    """Apply a segment to a visitor before serving the page.

    :param page: The page being served
    :type page: wagtail.core.models.Page
    :param request: The http request
    :type request: django.http.HttpRequest

    """
    adapter = get_segment_adapter(request)
    adapter.refresh()

    forced_segment = request.GET.get('segment', None)
    if request.user.is_superuser and forced_segment is not None:
        segment = models.Segment.objects.filter(pk=forced_segment).first()
        if segment:
            adapter.set_segments([segment])


class UserbarSegmentedLinkItem:
    def __init__(self, segment):
        self.segment = segment

    def render(self, request):
        return f"""<div class="wagtail-userbar__item">
            <a href="{request.path}?segment={self.segment.pk}"
                class="wagtail-action">
                    Show as segment: {self.segment.name}</a></div>"""


@hooks.register('construct_wagtail_userbar')
def add_segment_link_items(request, items):
    for item in models.Segment.objects.enabled():
        items.append(UserbarSegmentedLinkItem(item))
    return items


@hooks.register('before_serve_page')
def serve_variant(page, request, serve_args, serve_kwargs):
    """Apply a segment to a visitor before serving the page.

    :param page: The page being served
    :type page: wagtail.core.models.Page
    :param request: The http request
    :type request: django.http.HttpRequest
    :returns: A variant if one is available for the visitor's segment,
              otherwise the original page
    :rtype: wagtail.core.models.Page

    """
    user_segments = []
    if not isinstance(page, models.PersonalisablePageMixin):
        return

    adapter = get_segment_adapter(request)
    user_segments = adapter.get_segments()

    metadata = page.personalisation_metadata

    # If page is not canonical, don't serve it.
    if not metadata.is_canonical:
        raise Http404

    if user_segments:
        # TODO: This is never more then one page? (fix query count)
        metadata = metadata.metadata_for_segments(user_segments)
        if metadata:
            variant = metadata.first().variant.specific
            return variant.serve(request, *serve_args, **serve_kwargs)


@hooks.register('construct_explorer_page_queryset')
def dont_show_variant(parent_page, pages, request):
    return utils.exclude_variants(pages)


@hooks.register('register_page_listing_buttons')
def page_listing_variant_buttons(page, page_perms, is_parent=False):
    """Adds page listing buttons to personalisable pages. Shows variants for
    the page (if any) and a 'Create a new variant' button.

    """
    if not isinstance(page, models.PersonalisablePageMixin):
        return

    metadata = page.personalisation_metadata
    if metadata.is_canonical:
        yield ButtonWithDropdownFromHook(
            _('Variants'),
            hook_name='register_page_listing_variant_buttons',
            page=page,
            page_perms=page_perms,
            is_parent=is_parent,
            attrs={'target': '_blank', 'title': _('Create or edit a variant')},
            priority=100)


@hooks.register('register_page_listing_variant_buttons')
def page_listing_more_buttons(page, page_perms, is_parent=False):
    """Adds a 'more' button to personalisable pages allowing users to quickly
    create a new variant for the selected segment.

    """
    if not isinstance(page, models.PersonalisablePageMixin):
        return

    metadata = page.personalisation_metadata

    for vm in metadata.variants_metadata:
        yield Button('%s variant' % (vm.segment.name),
                     reverse('wagtailadmin_pages:edit', args=[vm.variant_id]),
                     attrs={"title": _('Edit this variant')},
                     classes=("icon", "icon-fa-pencil"),
                     priority=0)

    for segment in metadata.get_unused_segments():
        yield Button('%s variant' % (segment.name),
                     reverse('segment:copy_page', args=[page.pk, segment.pk]),
                     attrs={"title": _('Create this variant')},
                     classes=("icon", "icon-fa-plus"),
                     priority=100)

    yield Button(_('Create a new segment'),
                 reverse('wagtail_personalisation_segment_modeladmin_create'),
                 attrs={"title": _('Create a new segment')},
                 classes=("icon", "icon-fa-snowflake-o"),
                 priority=200)


class CorrectedPagesSummaryItem(PagesSummaryItem):
    def get_context(self):
        # Perform the same check as Wagtail to get the correct count.
        # Only correct the count when a root page is available to the user.
        # The `PagesSummaryItem` will return a page count of 0 otherwise.
        # https://github.com/wagtail/wagtail/blob/5c9ff23e229acabad406c42c4e13cbaea32e6c15/wagtail/admin/site_summary.py#L38
        context = super().get_context()
        root_page = context.get('root_page', None)
        if root_page:
            pages = utils.exclude_variants(
                Page.objects.descendant_of(root_page, inclusive=True))
            page_count = pages.count()

            if root_page.is_root():
                page_count -= 1

            context['total_pages'] = page_count

        return context


@hooks.register('construct_homepage_summary_items')
def add_corrected_pages_summary_panel(request, items):
    """Replaces the Pages summary panel to hide variants."""
    for index, item in enumerate(items):
        if item.__class__ is PagesSummaryItem:
            items[index] = CorrectedPagesSummaryItem(request)


class SegmentSummaryPanel(SummaryItem):
    """The segment summary panel showing the total amount of segments on the
    site and allowing quick access to the Segment dashboard.

    """
    order = 2000

    def render(self):
        segment_count = models.Segment.objects.count()
        target_url = reverse('wagtail_personalisation_segment_modeladmin_index')
        title = _("Segments")
        return mark_safe("""
            <li class="icon icon-fa-snowflake-o">
                <a href="{}"><span>{}</span>{}</a>
            </li>""".format(target_url, segment_count, title))


class PersonalisedPagesSummaryPanel(PagesSummaryItem):
    order = 2100

    def render(self):
        page_count = models.PersonalisablePageMetadata.objects.filter(
            segment__isnull=True).count()
        title = _("Personalised Page")
        return mark_safe("""
            <li class="icon icon-fa-file-o">
                <span>{}</span>{}{}
            </li>""".format(page_count, title, pluralize(page_count)))


class VariantPagesSummaryPanel(PagesSummaryItem):
    order = 2200

    def render(self):
        page_count = models.PersonalisablePageMetadata.objects.filter(
            segment__isnull=False).count()
        title = _("Variant")
        return mark_safe("""
                <li class="icon icon-fa-files-o">
                    <span>{}</span>{}{}
                </li>""".format(page_count, title, pluralize(page_count)))


@hooks.register('construct_homepage_summary_items')
def add_personalisation_summary_panels(request, items):
    """Adds a summary panel to the Wagtail dashboard showing the total amount
    of segments on the site and allowing quick access to the Segment
    dashboard.

    """
    items.append(SegmentSummaryPanel(request))
    items.append(PersonalisedPagesSummaryPanel(request))
    items.append(VariantPagesSummaryPanel(request))


@hooks.register('before_delete_page')
def delete_related_variants(request, page):
    if not isinstance(page, models.PersonalisablePageMixin) \
            or not page.personalisation_metadata.is_canonical:
        return
    # Get a list of related personalisation metadata for all the related
    # variants.
    variants_metadata = (
        page.personalisation_metadata.variants_metadata
                                     .select_related('variant')
    )
    next_url = get_valid_next_url_from_request(request)

    if request.method == 'POST':
        parent_id = page.get_parent().id
        variants_metadata = variants_metadata.select_related('variant')
        with transaction.atomic():
            for metadata in variants_metadata.iterator():
                # Call delete() on objects to trigger any signals or hooks.
                metadata.variant.delete()
            # Delete the page's main variant and the page itself.
            page.personalisation_metadata.delete()
            page.delete()
        msg = _("Page '{0}' and its variants deleted.")
        messages.success(
            request,
            msg.format(page.get_admin_display_title())
        )

        for fn in hooks.get_hooks('after_delete_page'):
            result = fn(request, page)
            if hasattr(result, 'status_code'):
                return result

        if next_url:
            return redirect(next_url)
        return redirect('wagtailadmin_explore', parent_id)

    return render(
        request,
        'wagtailadmin/pages/wagtail_personalisation/confirm_delete.html', {
            'page': page,
            'descendant_count': page.get_descendant_count(),
            'next': next_url,
            'variants': Page.objects.filter(
                pk__in=variants_metadata.values_list('variant_id')
            )
        }
    )



import random

from django import forms
from django.conf import settings
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models, transaction
from django.template.defaultfilters import slugify
from django.utils.encoding import python_2_unicode_compatible
from django.utils.functional import cached_property
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _
from modelcluster.models import ClusterableModel
import wagtail
from wagtail.admin.edit_handlers import (
    FieldPanel, FieldRowPanel, InlinePanel, MultiFieldPanel)
from wagtail.core.models import Page

from wagtail_personalisation.rules import AbstractBaseRule
from wagtail_personalisation.utils import count_active_days

from .forms import SegmentAdminForm


class SegmentQuerySet(models.QuerySet):
    def enabled(self):
        return self.filter(status=self.model.STATUS_ENABLED)


@python_2_unicode_compatible
class Segment(ClusterableModel):
    """The segment model."""
    STATUS_ENABLED = 'enabled'
    STATUS_DISABLED = 'disabled'

    STATUS_CHOICES = (
        (STATUS_ENABLED, _('Enabled')),
        (STATUS_DISABLED, _('Disabled')),
    )

    TYPE_DYNAMIC = 'dynamic'
    TYPE_STATIC = 'static'

    TYPE_CHOICES = (
        (TYPE_DYNAMIC, _('Dynamic')),
        (TYPE_STATIC, _('Static')),
    )

    name = models.CharField(max_length=255)
    create_date = models.DateTimeField(auto_now_add=True)
    edit_date = models.DateTimeField(auto_now=True)
    enable_date = models.DateTimeField(null=True, editable=False)
    disable_date = models.DateTimeField(null=True, editable=False)
    visit_count = models.PositiveIntegerField(default=0, editable=False)
    status = models.CharField(
        max_length=20, choices=STATUS_CHOICES, default=STATUS_ENABLED)
    persistent = models.BooleanField(
        default=False, help_text=_("Should the segment persist between visits?"))
    match_any = models.BooleanField(
        default=False,
        help_text=_("Should the segment match all the rules or just one of them?")
    )
    type = models.CharField(
        max_length=20,
        choices=TYPE_CHOICES,
        default=TYPE_DYNAMIC,
        help_text=mark_safe(_("""
            </br></br><strong>Dynamic:</strong> Users in this segment will change
            as more or less meet the rules specified in the segment.
            </br><strong>Static:</strong> If the segment contains only static
            compatible rules the segment will contain the members that pass
            those rules when the segment is created. Mixed static segments or
            those containing entirely non static compatible rules will be
            populated using the count variable.
        """))
    )
    count = models.PositiveSmallIntegerField(
        default=0,
        help_text=_(
            "If this number is set for a static segment users will be added to the "
            "set until the number is reached. After this no more users will be added."
        )
    )
    static_users = models.ManyToManyField(
        settings.AUTH_USER_MODEL,
    )
    excluded_users = models.ManyToManyField(
        settings.AUTH_USER_MODEL,
        help_text=_("Users that matched the rules but were excluded from the "
                    "segment for some reason e.g. randomisation"),
        related_name="excluded_segments"
    )

    matched_users_count = models.PositiveIntegerField(default=0, editable=False)
    matched_count_updated_at = models.DateTimeField(null=True, editable=False)

    randomisation_percent = models.PositiveSmallIntegerField(
        null=True, blank=True, default=None,
        help_text=_(
            "If this number is set each user matching the rules will "
            "have this percentage chance of being placed in the segment."
        ), validators=[
            MaxValueValidator(100),
            MinValueValidator(0)
        ])

    objects = SegmentQuerySet.as_manager()

    base_form_class = SegmentAdminForm

    def __init__(self, *args, **kwargs):
        Segment.panels = [
            MultiFieldPanel([
                FieldPanel('name', classname="title"),
                FieldRowPanel([
                    FieldPanel('status'),
                    FieldPanel('persistent'),
                ]),
                FieldPanel('match_any'),
                FieldPanel('type', widget=forms.RadioSelect),
                FieldPanel('count', classname='count_field'),
                FieldPanel('randomisation_percent', classname='percent_field'),
            ], heading="Segment"),
            MultiFieldPanel([
                InlinePanel(
                    "{}s".format(rule_model._meta.db_table),
                    label='{}{}'.format(
                        rule_model._meta.verbose_name,
                        ' ({})'.format(_('Static compatible')) if rule_model.static else ''
                    ),
                ) for rule_model in AbstractBaseRule.__subclasses__()
            ], heading=_("Rules")),
        ]

        super(Segment, self).__init__(*args, **kwargs)

    def __str__(self):
        return self.name

    @property
    def is_static(self):
        return self.type == self.TYPE_STATIC

    @classmethod
    def all_static(cls, rules):
        return all(rule.static for rule in rules)

    @property
    def all_rules_static(self):
        rules = self.get_rules()
        return rules and self.all_static(rules)

    @property
    def is_full(self):
        return self.static_users.count() >= self.count

    def encoded_name(self):
        """Return a string with a slug for the segment."""
        return slugify(self.name.lower())

    def get_active_days(self):
        """Return the amount of days the segment has been active."""
        return count_active_days(self.enable_date, self.disable_date)

    def get_used_pages(self):
        """Return the pages that have variants using this segment."""
        return PersonalisablePageMetadata.objects.filter(segment=self)

    def get_created_variants(self):
        """Return the variants using this segment."""
        return Page.objects.filter(_personalisable_page_metadata__segment=self)

    def get_rules(self):
        """Retrieve all rules in the segment."""
        segment_rules = []
        for rule_model in AbstractBaseRule.get_descendant_models():
            segment_rules.extend(
                rule_model._default_manager.filter(segment=self))

        return segment_rules

    def toggle(self, save=True):
        self.status = (
            self.STATUS_ENABLED if self.status == self.STATUS_DISABLED
            else self.STATUS_DISABLED)
        if save:
            self.save()

    def randomise_into_segment(self):
        """ Returns True if randomisation_percent is not set or it generates
        a random number less than the randomisation_percent
        This is so there is some randomisation in which users are added to the
        segment
        """
        if self.randomisation_percent is None:
            return True

        if random.randint(1, 100) <= self.randomisation_percent:
            return True
        return False


class PersonalisablePageMetadata(ClusterableModel):
    """The personalisable page model. Allows creation of variants with linked
    segments.

    """
    # Canonical pages should not ever be deleted if they have variants
    # because the variants will be orphaned.
    canonical_page = models.ForeignKey(
        Page, models.PROTECT, related_name='personalisable_canonical_metadata',
        null=True
    )

    # Delete metadata of the variant if its page gets deleted.
    variant = models.OneToOneField(
        Page, models.CASCADE, related_name='_personalisable_page_metadata',
        null=True
    )

    segment = models.ForeignKey(Segment, models.PROTECT, null=True,
                                related_name='page_metadata')

    @cached_property
    def has_variants(self):
        """Return a boolean indicating whether or not the personalisable page
        has variants.

        :returns: A boolean indicating whether or not the personalisable page
                  has variants.
        :rtype: bool

        """
        return self.variants_metadata.exists()

    @cached_property
    def variants_metadata(self):
        return (
            PersonalisablePageMetadata.objects
            .filter(canonical_page_id=self.canonical_page_id)
            .exclude(variant_id=self.variant_id)
            .exclude(variant_id=self.canonical_page_id))

    @cached_property
    def is_canonical(self):
        """Return a boolean indicating whether or not the personalisable page
        is a canonical page.

        :returns: A boolean indicating whether or not the personalisable
        page
                  is a canonical page.
        :rtype: bool

        """
        return self.canonical_page_id == self.variant_id

    def copy_for_segment(self, segment):
        page = self.canonical_page

        slug = "{}-{}".format(page.slug, segment.encoded_name())
        title = "{} ({})".format(page.title, segment.name)
        update_attrs = {
            'title': title,
            'slug': slug,
            'live': False,
        }

        with transaction.atomic():
            new_page = self.canonical_page.copy(
                update_attrs=update_attrs, copy_revisions=False)

            PersonalisablePageMetadata.objects.create(
                canonical_page=page,
                variant=new_page,
                segment=segment)
        return new_page

    def metadata_for_segments(self, segments):
        return (
            self.__class__.objects
            .filter(
                canonical_page_id=self.canonical_page_id,
                segment__in=segments))

    def get_unused_segments(self):
        if self.is_canonical:
            return (
                Segment.objects
                .exclude(page_metadata__canonical_page_id=self.canonical_page_id))
        return Segment.objects.none()


class PersonalisablePageMixin:
    """The personalisable page model. Allows creation of variants with linked
    segments.

    """

    @cached_property
    def personalisation_metadata(self):
        try:
            metadata = self._personalisable_page_metadata
        except AttributeError:
            metadata = PersonalisablePageMetadata.objects.create(
                canonical_page=self, variant=self)
        return metadata

    def get_sitemap_urls(self, request=None):
        # Do not generate sitemap entries for variants.
        if not self.personalisation_metadata.is_canonical:
            return []
        if wagtail.VERSION >= (2, 2):
            # Since Wagtail 2.2 you can pass request to the get_sitemap_urls
            # method.
            return super(PersonalisablePageMixin, self).get_sitemap_urls(
                request=request
            )
        return super(PersonalisablePageMixin, self).get_sitemap_urls()



from __future__ import absolute_import, unicode_literals

from django.conf import settings
from django.db.models import F
from django.utils.module_loading import import_string

from wagtail_personalisation.models import Segment
from wagtail_personalisation.rules import AbstractBaseRule
from wagtail_personalisation.utils import create_segment_dictionary


class BaseSegmentsAdapter:
    """Base segments adapter."""

    def __init__(self, request):
        """Prepare the request session for segment storage.

        :param request: The http request
        :type request: django.http.HttpRequest

        """
        self.request = request

    def setup(self):
        """Prepare the adapter for segment storage."""

    def get_segments(self):
        """Return the segments stored in the adapter storage."""

    def get_segment_by_id(self):
        """Return a single segment stored in the adapter storage."""

    def add(self):
        """Add a new segment to the adapter storage."""

    def refresh(self):
        """Refresh the segments stored in the adapter storage."""

    def _test_rules(self, rules, request, match_any=False):
        """Tests the provided rules to see if the request still belongs
        to a segment.
        :param rules: The rules to test for
        :type rules: list of wagtail_personalisation.rules
        :param request: The http request
        :type request: django.http.HttpRequest
        :param match_any: Whether all rules need to match, or any
        :type match_any: bool
        :returns: A boolean indicating the segment matches the request
        :rtype: bool
        """
        if not rules:
            return False
        if match_any:
            return any(rule.test_user(request) for rule in rules)
        return all(rule.test_user(request) for rule in rules)

    class Meta:
        abstract = True


class SessionSegmentsAdapter(BaseSegmentsAdapter):
    """Segment adapter that uses Django's session backend."""

    def __init__(self, request):
        super(SessionSegmentsAdapter, self).__init__(request)
        self.request.session.setdefault('segments', [])
        self._segment_cache = None

    def _segments(self, ids=None):
        if not ids:
            ids = []
        segments = (
            Segment.objects
            .enabled()
            .filter(persistent=True)
            .filter(pk__in=ids)
        )
        return segments

    def get_segments(self, key="segments"):
        """Return the persistent segments stored in the request session.

        :param key: The key under which the segments are stored
        :type key: String
        :returns: The segments in the request session
        :rtype: list of wagtail_personalisation.models.Segment or empty list

        """
        if key == "segments" and self._segment_cache is not None:
            return self._segment_cache

        if key not in self.request.session:
            return []
        raw_segments = self.request.session[key]
        segment_ids = [segment['id'] for segment in raw_segments]

        segments = self._segments(ids=segment_ids)

        result = list(segments)
        if key == "segments":
            self._segment_cache = result
        return result

    def set_segments(self, segments, key="segments"):
        """Set the currently active segments

        :param segments: The segments to set for the current request
        :type segments: list of wagtail_personalisation.models.Segment
        :param key: The key under which to store the segments. Optional
        :type key: String

        """
        cache_segments = []
        serialized_segments = []
        segment_ids = set()
        for segment in segments:
            serialized = create_segment_dictionary(segment)
            if serialized['id'] in segment_ids:
                continue

            cache_segments.append(segment)
            serialized_segments.append(serialized)
            segment_ids.add(segment.pk)

        self.request.session[key] = serialized_segments
        if key == "segments":
            self._segment_cache = cache_segments

    def get_segment_by_id(self, segment_id):
        """Find and return a single segment from the request session.

        :param segment_id: The primary key of the segment
        :type segment_id: int
        :returns: The matching segment
        :rtype: wagtail_personalisation.models.Segment or None

        """
        segments = self._segments(ids=[segment_id])
        if segments.exists():
            return segments.get()

    def add_page_visit(self, page):
        """Mark the page as visited by the user"""
        visit_count = self.request.session.setdefault('visit_count', [])
        page_visits = [visit for visit in visit_count if visit['id'] == page.pk]

        if page_visits:
            for page_visit in page_visits:
                page_visit['count'] += 1
                page_visit['path'] = page.url_path if page else self.request.path
            self.request.session.modified = True
        else:
            visit_count.append({
                'slug': page.slug,
                'id': page.pk,
                'path': page.url_path if page else self.request.path,
                'count': 1,
            })

    def get_visit_count(self, page=None):
        """Return the number of visits on the current request or given page"""
        path = page.url_path if page else self.request.path
        visit_count = self.request.session.setdefault('visit_count', [])
        for visit in visit_count:
            if visit['path'] == path:
                return visit['count']
        return 0

    def update_visit_count(self):
        """Update the visit count for all segments in the request session."""
        segments = self.request.session['segments']
        segment_pks = [s['id'] for s in segments]

        # Update counts
        (Segment.objects
            .enabled()
            .filter(pk__in=segment_pks)
            .update(visit_count=F('visit_count') + 1))

    def refresh(self):
        """Retrieve the request session segments and verify whether or not they
        still apply to the requesting visitor.

        """
        enabled_segments = Segment.objects.enabled()
        rule_models = AbstractBaseRule.get_descendant_models()

        current_segments = self.get_segments()
        excluded_segments = self.get_segments("excluded_segments")
        current_segments = list(
            set(current_segments) - set(excluded_segments)
        )

        # Run tests on all remaining enabled segments to verify applicability.
        additional_segments = []
        for segment in enabled_segments:
            if segment.is_static and segment.static_users.filter(id=self.request.user.id).exists():
                additional_segments.append(segment)
            elif (segment.excluded_users.filter(id=self.request.user.id).exists() or
                    segment in excluded_segments):
                continue
            elif not segment.is_static or not segment.is_full:
                segment_rules = []
                for rule_model in rule_models:
                    segment_rules.extend(rule_model.objects.filter(segment=segment))

                result = self._test_rules(segment_rules, self.request,
                                          match_any=segment.match_any)

                if result and segment.randomise_into_segment():
                    if segment.is_static and not segment.is_full:
                        if self.request.user.is_authenticated:
                            segment.static_users.add(self.request.user)
                    additional_segments.append(segment)
                elif result:
                    if segment.is_static and self.request.user.is_authenticated:
                        segment.excluded_users.add(self.request.user)
                    else:
                        excluded_segments += [segment]

        self.set_segments(current_segments + additional_segments)
        self.set_segments(excluded_segments, "excluded_segments")
        self.update_visit_count()


SEGMENT_ADAPTER_CLASS = import_string(getattr(
    settings,
    'PERSONALISATION_SEGMENTS_ADAPTER',
    'wagtail_personalisation.adapters.SessionSegmentsAdapter'))


def get_segment_adapter(request):
    """Return the Segment Adapter for the given request"""
    if not hasattr(request, 'segment_adapter'):
        request.segment_adapter = SEGMENT_ADAPTER_CLASS(request)
    return request.segment_adapter



# Generated by Django 2.0.6 on 2018-08-10 14:39

from django.db import migrations, models
import django.db.models.deletion
import modelcluster.fields


class Migration(migrations.Migration):

    dependencies = [
        ('wagtail_personalisation', '0023_personalisablepagemetadata_variant_cascade'),
    ]

    operations = [
        migrations.CreateModel(
            name='OriginCountryRule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('country', models.CharField(choices=[('aw', 'Aruba'), ('af', 'Afghanistan'), ('ao', 'Angola'), ('ai', 'Anguilla'), ('ax', 'Åland Islands'), ('al', 'Albania'), ('ad', 'Andorra'), ('ae', 'United Arab Emirates'), ('ar', 'Argentina'), ('am', 'Armenia'), ('as', 'American Samoa'), ('aq', 'Antarctica'), ('tf', 'French Southern Territories'), ('ag', 'Antigua and Barbuda'), ('au', 'Australia'), ('at', 'Austria'), ('az', 'Azerbaijan'), ('bi', 'Burundi'), ('be', 'Belgium'), ('bj', 'Benin'), ('bq', 'Bonaire, Sint Eustatius and Saba'), ('bf', 'Burkina Faso'), ('bd', 'Bangladesh'), ('bg', 'Bulgaria'), ('bh', 'Bahrain'), ('bs', 'Bahamas'), ('ba', 'Bosnia and Herzegovina'), ('bl', 'Saint Barthélemy'), ('by', 'Belarus'), ('bz', 'Belize'), ('bm', 'Bermuda'), ('bo', 'Bolivia, Plurinational State of'), ('br', 'Brazil'), ('bb', 'Barbados'), ('bn', 'Brunei Darussalam'), ('bt', 'Bhutan'), ('bv', 'Bouvet Island'), ('bw', 'Botswana'), ('cf', 'Central African Republic'), ('ca', 'Canada'), ('cc', 'Cocos (Keeling) Islands'), ('ch', 'Switzerland'), ('cl', 'Chile'), ('cn', 'China'), ('ci', "Côte d'Ivoire"), ('cm', 'Cameroon'), ('cd', 'Congo, The Democratic Republic of the'), ('cg', 'Congo'), ('ck', 'Cook Islands'), ('co', 'Colombia'), ('km', 'Comoros'), ('cv', 'Cabo Verde'), ('cr', 'Costa Rica'), ('cu', 'Cuba'), ('cw', 'Curaçao'), ('cx', 'Christmas Island'), ('ky', 'Cayman Islands'), ('cy', 'Cyprus'), ('cz', 'Czechia'), ('de', 'Germany'), ('dj', 'Djibouti'), ('dm', 'Dominica'), ('dk', 'Denmark'), ('do', 'Dominican Republic'), ('dz', 'Algeria'), ('ec', 'Ecuador'), ('eg', 'Egypt'), ('er', 'Eritrea'), ('eh', 'Western Sahara'), ('es', 'Spain'), ('ee', 'Estonia'), ('et', 'Ethiopia'), ('fi', 'Finland'), ('fj', 'Fiji'), ('fk', 'Falkland Islands (Malvinas)'), ('fr', 'France'), ('fo', 'Faroe Islands'), ('fm', 'Micronesia, Federated States of'), ('ga', 'Gabon'), ('gb', 'United Kingdom'), ('ge', 'Georgia'), ('gg', 'Guernsey'), ('gh', 'Ghana'), ('gi', 'Gibraltar'), ('gn', 'Guinea'), ('gp', 'Guadeloupe'), ('gm', 'Gambia'), ('gw', 'Guinea-Bissau'), ('gq', 'Equatorial Guinea'), ('gr', 'Greece'), ('gd', 'Grenada'), ('gl', 'Greenland'), ('gt', 'Guatemala'), ('gf', 'French Guiana'), ('gu', 'Guam'), ('gy', 'Guyana'), ('hk', 'Hong Kong'), ('hm', 'Heard Island and McDonald Islands'), ('hn', 'Honduras'), ('hr', 'Croatia'), ('ht', 'Haiti'), ('hu', 'Hungary'), ('id', 'Indonesia'), ('im', 'Isle of Man'), ('in', 'India'), ('io', 'British Indian Ocean Territory'), ('ie', 'Ireland'), ('ir', 'Iran, Islamic Republic of'), ('iq', 'Iraq'), ('is', 'Iceland'), ('il', 'Israel'), ('it', 'Italy'), ('jm', 'Jamaica'), ('je', 'Jersey'), ('jo', 'Jordan'), ('jp', 'Japan'), ('kz', 'Kazakhstan'), ('ke', 'Kenya'), ('kg', 'Kyrgyzstan'), ('kh', 'Cambodia'), ('ki', 'Kiribati'), ('kn', 'Saint Kitts and Nevis'), ('kr', 'Korea, Republic of'), ('kw', 'Kuwait'), ('la', "Lao People's Democratic Republic"), ('lb', 'Lebanon'), ('lr', 'Liberia'), ('ly', 'Libya'), ('lc', 'Saint Lucia'), ('li', 'Liechtenstein'), ('lk', 'Sri Lanka'), ('ls', 'Lesotho'), ('lt', 'Lithuania'), ('lu', 'Luxembourg'), ('lv', 'Latvia'), ('mo', 'Macao'), ('mf', 'Saint Martin (French part)'), ('ma', 'Morocco'), ('mc', 'Monaco'), ('md', 'Moldova, Republic of'), ('mg', 'Madagascar'), ('mv', 'Maldives'), ('mx', 'Mexico'), ('mh', 'Marshall Islands'), ('mk', 'Macedonia, Republic of'), ('ml', 'Mali'), ('mt', 'Malta'), ('mm', 'Myanmar'), ('me', 'Montenegro'), ('mn', 'Mongolia'), ('mp', 'Northern Mariana Islands'), ('mz', 'Mozambique'), ('mr', 'Mauritania'), ('ms', 'Montserrat'), ('mq', 'Martinique'), ('mu', 'Mauritius'), ('mw', 'Malawi'), ('my', 'Malaysia'), ('yt', 'Mayotte'), ('na', 'Namibia'), ('nc', 'New Caledonia'), ('ne', 'Niger'), ('nf', 'Norfolk Island'), ('ng', 'Nigeria'), ('ni', 'Nicaragua'), ('nu', 'Niue'), ('nl', 'Netherlands'), ('no', 'Norway'), ('np', 'Nepal'), ('nr', 'Nauru'), ('nz', 'New Zealand'), ('om', 'Oman'), ('pk', 'Pakistan'), ('pa', 'Panama'), ('pn', 'Pitcairn'), ('pe', 'Peru'), ('ph', 'Philippines'), ('pw', 'Palau'), ('pg', 'Papua New Guinea'), ('pl', 'Poland'), ('pr', 'Puerto Rico'), ('kp', "Korea, Democratic People's Republic of"), ('pt', 'Portugal'), ('py', 'Paraguay'), ('ps', 'Palestine, State of'), ('pf', 'French Polynesia'), ('qa', 'Qatar'), ('re', 'Réunion'), ('ro', 'Romania'), ('ru', 'Russian Federation'), ('rw', 'Rwanda'), ('sa', 'Saudi Arabia'), ('sd', 'Sudan'), ('sn', 'Senegal'), ('sg', 'Singapore'), ('gs', 'South Georgia and the South Sandwich Islands'), ('sh', 'Saint Helena, Ascension and Tristan da Cunha'), ('sj', 'Svalbard and Jan Mayen'), ('sb', 'Solomon Islands'), ('sl', 'Sierra Leone'), ('sv', 'El Salvador'), ('sm', 'San Marino'), ('so', 'Somalia'), ('pm', 'Saint Pierre and Miquelon'), ('rs', 'Serbia'), ('ss', 'South Sudan'), ('st', 'Sao Tome and Principe'), ('sr', 'Suriname'), ('sk', 'Slovakia'), ('si', 'Slovenia'), ('se', 'Sweden'), ('sz', 'Swaziland'), ('sx', 'Sint Maarten (Dutch part)'), ('sc', 'Seychelles'), ('sy', 'Syrian Arab Republic'), ('tc', 'Turks and Caicos Islands'), ('td', 'Chad'), ('tg', 'Togo'), ('th', 'Thailand'), ('tj', 'Tajikistan'), ('tk', 'Tokelau'), ('tm', 'Turkmenistan'), ('tl', 'Timor-Leste'), ('to', 'Tonga'), ('tt', 'Trinidad and Tobago'), ('tn', 'Tunisia'), ('tr', 'Turkey'), ('tv', 'Tuvalu'), ('tw', 'Taiwan, Province of China'), ('tz', 'Tanzania, United Republic of'), ('ug', 'Uganda'), ('ua', 'Ukraine'), ('um', 'United States Minor Outlying Islands'), ('uy', 'Uruguay'), ('us', 'United States'), ('uz', 'Uzbekistan'), ('va', 'Holy See (Vatican City State)'), ('vc', 'Saint Vincent and the Grenadines'), ('ve', 'Venezuela, Bolivarian Republic of'), ('vg', 'Virgin Islands, British'), ('vi', 'Virgin Islands, U.S.'), ('vn', 'Viet Nam'), ('vu', 'Vanuatu'), ('wf', 'Wallis and Futuna'), ('ws', 'Samoa'), ('ye', 'Yemen'), ('za', 'South Africa'), ('zm', 'Zambia'), ('zw', 'Zimbabwe')], help_text='Select origin country of the request that this rule will match against. This rule will only work if you use Cloudflare or CloudFront IP geolocation or if GeoIP2 module is configured.', max_length=2)),
                ('segment', modelcluster.fields.ParentalKey(on_delete=django.db.models.deletion.CASCADE, related_name='wagtail_personalisation_origincountryrules', to='wagtail_personalisation.Segment')),
            ],
            options={
                'verbose_name': 'origin country rule',
            },
        ),
    ]



import re
from setuptools import find_packages, setup

install_requires = [
    'wagtail>=2.0',
    'user-agents>=1.1.0',
    'wagtailfontawesome>=1.1.3',
    'pycountry',
]

tests_require = [
    'factory_boy==2.8.1',
    'flake8-blind-except',
    'flake8-debugger',
    'flake8-imports',
    'flake8',
    'freezegun==0.3.8',
    'pytest-cov==2.5.1',
    'pytest-django==3.1.2',
    'pytest-pythonpath==0.7.2',
    'pytest-sugar==0.9.1',
    'pytest==3.4.2',
    'wagtail_factories==1.0.0',
    'pytest-mock==1.6.3',
]

docs_require = [
    'sphinx>=1.7.6',
    'sphinx_rtd_theme>=0.4.0',
]

with open('README.rst') as fh:
    long_description = re.sub(
        '^.. start-no-pypi.*^.. end-no-pypi', '', fh.read(), flags=re.M | re.S)

setup(
    name='wagtail-personalisation',
    version='0.12.1',
    description='A Wagtail add-on for showing personalized content',
    author='Lab Digital BV and others',
    author_email='opensource@labdigital.nl',
    url='https://labdigital.nl/',
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require={
        'docs': docs_require,
        'test': tests_require,
    },
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    license='MIT',
    long_description=long_description,
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Framework :: Django',
        'Framework :: Django :: 2.0',
        'Topic :: Internet :: WWW/HTTP :: Site Management',
    ],
)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-05-31 14:28
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wagtailcore', '0001_initial'),
        ('wagtail_personalisation', '0010_auto_20170531_1101'),
    ]

    operations = [
        migrations.CreateModel(
            name='PersonalisablePageMetadata',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_segmented', models.BooleanField(default=False)),
                ('canonical_page', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='personalisable_canonical_metadata', to='wagtailcore.Page')),
                ('segment', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='page_metadata', to='wagtail_personalisation.Segment')),
                ('variant', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='_personalisable_page_metadata', to='wagtailcore.Page')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



from __future__ import absolute_import, unicode_literals
import logging

import re
from datetime import datetime
from importlib import import_module

import pycountry
from django.apps import apps
from django.conf import settings
from django.contrib.sessions.models import Session
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.template.defaultfilters import slugify
from django.test.client import RequestFactory
from django.utils.encoding import force_text, python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _
from modelcluster.fields import ParentalKey
from user_agents import parse
from wagtail.admin.edit_handlers import (
    FieldPanel, FieldRowPanel, PageChooserPanel)

from wagtail_personalisation.utils import get_client_ip

SessionStore = import_module(settings.SESSION_ENGINE).SessionStore

logger = logging.getLogger(__name__)


def get_geoip_module():
    try:
        from django.contrib.gis.geoip2 import GeoIP2
        return GeoIP2
    except ImportError:
        logger.exception(
            'GeoIP module is disabled. To use GeoIP for the origin\n'
            'country personaliastion rule please set it up as per '
            'documentation:\n'
            'https://docs.djangoproject.com/en/stable/ref/contrib/gis/'
            'geoip2/.\n'
            'Wagtail-personalisation also works with Cloudflare and\n'
            'CloudFront country detection, so you should not see this\n'
            'warning if you use one of those.')


@python_2_unicode_compatible
class AbstractBaseRule(models.Model):
    """Base for creating rules to segment users with."""
    icon = 'fa-circle-o'
    static = False

    segment = ParentalKey(
        'wagtail_personalisation.Segment',
        related_name="%(app_label)s_%(class)ss",
    )

    class Meta:
        abstract = True
        verbose_name = 'Abstract segmentation rule'

    def __str__(self):
        return force_text(self._meta.verbose_name)

    def test_user(self):
        """Test if the user matches this rule."""
        return True

    def encoded_name(self):
        """Return a string with a slug for the rule."""
        return slugify(force_text(self).lower())

    def description(self):
        """Return a description explaining the functionality of the rule.
        Used in the segmentation dashboard.

        :returns: A dict containing a title and a value
        :rtype: dict

        """
        description = {
            'title': _('Abstract segmentation rule'),
            'value': '',
        }

        return description

    @classmethod
    def get_descendant_models(cls):
        return [model for model in apps.get_models()
                if issubclass(model, AbstractBaseRule)]


class TimeRule(AbstractBaseRule):
    """Time rule to segment users based on a start and end time.

    Matches when the time a request is made falls between the
    set start time and end time.

    """
    icon = 'fa-clock-o'

    start_time = models.TimeField(_("Starting time"))
    end_time = models.TimeField(_("Ending time"))

    panels = [
        FieldRowPanel([
            FieldPanel('start_time'),
            FieldPanel('end_time'),
        ]),
    ]

    class Meta:
        verbose_name = _('Time Rule')

    def test_user(self, request=None):
        return self.start_time <= datetime.now().time() <= self.end_time

    def description(self):
        return {
            'title': _('These users visit between'),
            'value': _('{} and {}').format(
                self.start_time.strftime("%H:%M"),
                self.end_time.strftime("%H:%M")
            ),
        }


class DayRule(AbstractBaseRule):
    """Day rule to segment users based on the day(s) of a visit.

    Matches when the day a request is made matches with the days
    set in the rule.

    """
    icon = 'fa-calendar-check-o'

    mon = models.BooleanField(_("Monday"), default=False)
    tue = models.BooleanField(_("Tuesday"), default=False)
    wed = models.BooleanField(_("Wednesday"), default=False)
    thu = models.BooleanField(_("Thursday"), default=False)
    fri = models.BooleanField(_("Friday"), default=False)
    sat = models.BooleanField(_("Saturday"), default=False)
    sun = models.BooleanField(_("Sunday"), default=False)

    panels = [
        FieldPanel('mon'),
        FieldPanel('tue'),
        FieldPanel('wed'),
        FieldPanel('thu'),
        FieldPanel('fri'),
        FieldPanel('sat'),
        FieldPanel('sun'),
    ]

    class Meta:
        verbose_name = _('Day Rule')

    def test_user(self, request=None):
        return [self.mon, self.tue, self.wed, self.thu,
                self.fri, self.sat, self.sun][datetime.today().weekday()]

    def description(self):
        days = (
            ('mon', self.mon), ('tue', self.tue), ('wed', self.wed),
            ('thu', self.thu), ('fri', self.fri), ('sat', self.sat),
            ('sun', self.sun),
        )

        chosen_days = [day_name for day_name, chosen in days if chosen]

        return {
            'title': _('These users visit on'),
            'value': ", ".join([day for day in chosen_days]).title(),
        }


class ReferralRule(AbstractBaseRule):
    """Referral rule to segment users based on a regex test.

    Matches when the referral header in a request matches with
    the set regex test.

    """
    icon = 'fa-globe'

    regex_string = models.TextField(
        _("Regular expression to match the referrer"))

    panels = [
        FieldPanel('regex_string'),
    ]

    class Meta:
        verbose_name = _('Referral Rule')

    def test_user(self, request):
        pattern = re.compile(self.regex_string)

        if 'HTTP_REFERER' in request.META:
            referer = request.META['HTTP_REFERER']
            if pattern.search(referer):
                return True
        return False

    def description(self):
        return {
            'title': _('These visits originate from'),
            'value': self.regex_string,
            'code': True
        }


class VisitCountRule(AbstractBaseRule):
    """Visit count rule to segment users based on amount of visits to a
    specified page.

    Matches when the operator and count validate True
    when visiting the set page.

    """
    icon = 'fa-calculator'
    static = True

    OPERATOR_CHOICES = (
        ('more_than', _("More than")),
        ('less_than', _("Less than")),
        ('equal_to', _("Equal to")),
    )
    operator = models.CharField(max_length=20,
                                choices=OPERATOR_CHOICES, default="more_than")
    count = models.PositiveSmallIntegerField(default=0, null=True)
    counted_page = models.ForeignKey(
        'wagtailcore.Page',
        null=False,
        blank=False,
        on_delete=models.CASCADE,
        related_name='+',
    )

    panels = [
        PageChooserPanel('counted_page'),
        FieldRowPanel([
            FieldPanel('operator'),
            FieldPanel('count'),
        ]),
    ]

    class Meta:
        verbose_name = _('Visit count Rule')

    def _get_user_session(self, user):
        sessions = Session.objects.iterator()
        for session in sessions:
            session_data = session.get_decoded()
            if session_data.get('_auth_user_id') == str(user.id):
                return SessionStore(session_key=session.session_key)
        return SessionStore()

    def test_user(self, request, user=None):
        # Local import for cyclic import
        from wagtail_personalisation.adapters import (
            get_segment_adapter, SessionSegmentsAdapter, SEGMENT_ADAPTER_CLASS)

        # Django formsets don't honour 'required' fields so check rule is valid
        try:
            self.counted_page
        except ObjectDoesNotExist:
            return False

        if user:
            # Create a fake request so we can use the adapter
            request = RequestFactory().get('/')
            request.user = user

            # If we're using the session adapter check for an active session
            if SEGMENT_ADAPTER_CLASS == SessionSegmentsAdapter:
                request.session = self._get_user_session(user)
            else:
                request.session = SessionStore()

        elif not request:
            # Return false if we don't have a user or a request
            return False

        operator = self.operator
        segment_count = self.count

        adapter = get_segment_adapter(request)

        visit_count = adapter.get_visit_count(self.counted_page)
        if visit_count and operator == "more_than":
            if visit_count > segment_count:
                return True
        elif visit_count and operator == "less_than":
            if visit_count < segment_count:
                return True
        elif visit_count and operator == "equal_to":
            if visit_count == segment_count:
                return True
        return False

    def description(self):
        return {
            'title': _('These users visited {}').format(
                self.counted_page
            ),
            'value': _('{} {} times').format(
                self.get_operator_display(),
                self.count
            ),
        }

    def get_column_header(self):
        return "Visit count - %s" % self.counted_page

    def get_user_info_string(self, user):
        # Local import for cyclic import
        from wagtail_personalisation.adapters import (
            get_segment_adapter, SessionSegmentsAdapter, SEGMENT_ADAPTER_CLASS)

        # Create a fake request so we can use the adapter
        request = RequestFactory().get('/')
        request.user = user

        # If we're using the session adapter check for an active session
        if SEGMENT_ADAPTER_CLASS == SessionSegmentsAdapter:
            request.session = self._get_user_session(user)
        else:
            request.session = SessionStore()

        adapter = get_segment_adapter(request)
        visit_count = adapter.get_visit_count(self.counted_page)
        return str(visit_count)


class QueryRule(AbstractBaseRule):
    """Query rule to segment users based on matching queries.

    Matches when both the set parameter and value match with one
    present in the request query.

    """
    icon = 'fa-link'

    parameter = models.SlugField(_("The query parameter to search for"),
                                 max_length=20)
    value = models.SlugField(_("The value of the parameter to match"),
                             max_length=20)

    panels = [
        FieldPanel('parameter'),
        FieldPanel('value'),
    ]

    class Meta:
        verbose_name = _('Query Rule')

    def test_user(self, request):
        return request.GET.get(self.parameter, '') == self.value

    def description(self):
        return {
            'title': _('These users used a URL with the query'),
            'value': _('?{}={}').format(
                self.parameter,
                self.value
            ),
            'code': True
        }


class DeviceRule(AbstractBaseRule):
    """Device rule to segment users based on matching devices.

    Matches when the set device type matches with the one present
    in the request user agent headers.

    """
    icon = 'fa-tablet'

    mobile = models.BooleanField(_("Mobile phone"), default=False)
    tablet = models.BooleanField(_("Tablet"), default=False)
    desktop = models.BooleanField(_("Desktop"), default=False)

    panels = [
        FieldPanel('mobile'),
        FieldPanel('tablet'),
        FieldPanel('desktop'),
    ]

    class Meta:
        verbose_name = _('Device Rule')

    def test_user(self, request=None):
        ua_header = request.META['HTTP_USER_AGENT']
        user_agent = parse(ua_header)

        if user_agent.is_mobile:
            return self.mobile
        if user_agent.is_tablet:
            return self.tablet
        if user_agent.is_pc:
            return self.desktop
        return False


class UserIsLoggedInRule(AbstractBaseRule):
    """User is logged in rule to segment users based on their authentication
    status.

    Matches when the user is authenticated.

    """
    icon = 'fa-user'

    is_logged_in = models.BooleanField(default=False)

    panels = [
        FieldPanel('is_logged_in'),
    ]

    class Meta:
        verbose_name = _('Logged in Rule')

    def test_user(self, request=None):
        return request.user.is_authenticated() == self.is_logged_in

    def description(self):
        return {
            'title': _('These visitors are'),
            'value': _('Logged in') if self.is_logged_in else _('Not logged in'),
        }


COUNTRY_CHOICES = [(country.alpha_2.lower(), country.name)
                   for country in pycountry.countries]


class OriginCountryRule(AbstractBaseRule):
    """
    Test user against the country or origin of their request.

    Using this rule requires setting up GeoIP2 on Django or using
    CloudFlare or CloudFront geolocation detection.
    """
    country = models.CharField(
        max_length=2, choices=COUNTRY_CHOICES,
        help_text=_("Select origin country of the request that this rule will "
                    "match against. This rule will only work if you use "
                    "Cloudflare or CloudFront IP geolocation or if GeoIP2 "
                    "module is configured.")
    )

    class Meta:
        verbose_name = _("origin country rule")

    def get_cloudflare_country(self, request):
        """
        Get country code that has been detected by Cloudflare.

        Guide to the functionality:
        https://support.cloudflare.com/hc/en-us/articles/200168236-What-does-Cloudflare-IP-Geolocation-do-
        """
        try:
            return request.META['HTTP_CF_IPCOUNTRY'].lower()
        except KeyError:
            pass

    def get_cloudfront_country(self, request):
        try:
            return request.META['HTTP_CLOUDFRONT_VIEWER_COUNTRY'].lower()
        except KeyError:
            pass

    def get_geoip_country(self, request):
        GeoIP2 = get_geoip_module()
        if GeoIP2 is None:
            return False
        return GeoIP2().country_code(get_client_ip(request)).lower()

    def get_country(self, request):
        # Prioritise CloudFlare and CloudFront country detection over GeoIP.
        functions = (
            self.get_cloudflare_country,
            self.get_cloudfront_country,
            self.get_geoip_country,
        )
        for function in functions:
            result = function(request)
            if result:
                return result

    def test_user(self, request=None):
        return (self.get_country(request) or '') == self.country.lower()



from django.template import Library

from wagtail_personalisation.utils import count_active_days

register = Library()


@register.filter(name='days_since')
def active_days(enable_date, disable_date):
    return count_active_days(enable_date, disable_date)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.8 on 2018-01-31 16:12
from __future__ import unicode_literals

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wagtail_personalisation', '0016_auto_20180125_0918'),
    ]

    operations = [
        migrations.AddField(
            model_name='segment',
            name='randomisation_percent',
            field=models.PositiveSmallIntegerField(blank=True, default=None, help_text='If this number is set each user matching the rules will have this percentage chance of being placed in the segment.', null=True, validators=[django.core.validators.MaxValueValidator(100), django.core.validators.MinValueValidator(0)]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2016-12-22 13:23
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wagtail_personalisation', '0005_userisloggedinrule'),
    ]

    operations = [
        migrations.AddField(
            model_name='segment',
            name='match_any',
            field=models.BooleanField(default=False, help_text='Should the segment match all the rules or just one of them?'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.6 on 2017-10-17 11:18
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('sessions', '0001_initial'),
        ('wagtail_personalisation', '0012_remove_personalisablepagemetadata_is_segmented'),
    ]

    operations = [
        migrations.AddField(
            model_name='segment',
            name='count',
            field=models.PositiveSmallIntegerField(default=0, help_text='If this number is set for a static segment users will be added to the set until the number is reached. After this no more users will be added.'),
        ),
        migrations.AddField(
            model_name='segment',
            name='sessions',
            field=models.ManyToManyField(to='sessions.Session'),
        ),
        migrations.AddField(
            model_name='segment',
            name='type',
            field=models.CharField(choices=[('dynamic', 'Dynamic'), ('static', 'Static')], default='dynamic', help_text='\n            </br></br><strong>Dynamic:</strong> Users in this segment will change\n            as more or less meet the rules specified in the segment.\n            </br><strong>Static:</strong> If the segment contains only static\n            compatible rules the segment will contain the members that pass\n            those rules when the segment is created. Mixed static segments or\n            those containing entirely non static compatible rules will be\n            populated using the count variable.\n        ', max_length=20),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-01-10 14:35
from __future__ import unicode_literals

import django.db.models.deletion
import modelcluster.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wagtail_personalisation', '0006_segment_match_any'),
    ]

    operations = [
        migrations.CreateModel(
            name='DayRule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('mon', models.BooleanField(default=False, verbose_name='Monday')),
                ('tue', models.BooleanField(default=False, verbose_name='Tuesday')),
                ('wed', models.BooleanField(default=False, verbose_name='Wednesday')),
                ('thu', models.BooleanField(default=False, verbose_name='Thursday')),
                ('fri', models.BooleanField(default=False, verbose_name='Friday')),
                ('sat', models.BooleanField(default=False, verbose_name='Saturday')),
                ('sun', models.BooleanField(default=False, verbose_name='Sunday')),
                ('segment', modelcluster.fields.ParentalKey(on_delete=django.db.models.deletion.CASCADE, related_name='wagtail_personalisation_dayrule_related', related_query_name='wagtail_personalisation_dayrules', to='wagtail_personalisation.Segment')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



from __future__ import absolute_import, unicode_literals

import csv

from django import forms
from django.core.exceptions import PermissionDenied
from django.db import transaction
from django.http import (
    HttpResponse, HttpResponseForbidden, HttpResponseRedirect)
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _
from wagtail.contrib.modeladmin.options import ModelAdmin, modeladmin_register
from wagtail.contrib.modeladmin.views import DeleteView, IndexView
from wagtail.core.models import Page

from wagtail_personalisation.models import Segment
from wagtail_personalisation.utils import can_delete_pages


class SegmentModelIndexView(IndexView):
    """Placeholder for additional list functionality."""
    pass


class SegmentModelDashboardView(IndexView):
    """Additional dashboard functionality."""
    def media(self):
        return forms.Media(
            css={'all': ['css/dashboard.css']},
            js=['js/commons.js', 'js/dashboard.js']
        )

    def get_template_names(self):
        return [
            'modeladmin/wagtail_personalisation/segment/dashboard.html',
            'modeladmin/index.html'
        ]


class SegmentModelDeleteView(DeleteView):
    def get_affected_page_objects(self):
        return Page.objects.filter(pk__in=(
            self.instance.get_used_pages().values_list('variant_id', flat=True)
        ))

    def get_template_names(self):
        return [
            'modeladmin/wagtail_personalisation/segment/delete.html',
            'modeladmin/delete.html',
        ]

    def delete_instance(self):
        page_variants = self.get_affected_page_objects()
        if not can_delete_pages(page_variants, self.request.user):
            raise PermissionDenied(
                'User has no permission to delete variant page objects.'
            )
        # Deleting page objects triggers deletion of the personalisation
        # metadata too because of models.CASCADE.
        with transaction.atomic():
            for variant in page_variants.iterator():
                # Delete each one separately so signals are called.
                variant.delete()
            super().delete_instance()

    def post(self, request, *args, **kwargs):
        if not can_delete_pages(self.get_affected_page_objects(),
                                self.request.user):
            context = self.get_context_data(
                cannot_delete_page_variants_error=True,
            )
            return self.render_to_response(context)
        return super().post(request, *args, **kwargs)


@modeladmin_register
class SegmentModelAdmin(ModelAdmin):
    """The model admin for the Segments administration interface."""
    model = Segment
    index_view_class = SegmentModelIndexView
    dashboard_view_class = SegmentModelDashboardView
    delete_view_class = SegmentModelDeleteView
    menu_icon = 'fa-snowflake-o'
    add_to_settings_menu = False
    list_display = ('name', 'persistent', 'match_any', 'status',
                    'page_count', 'variant_count', 'statistics')
    index_view_extra_js = ['js/commons.js', 'js/index.js']
    index_view_extra_css = ['css/index.css']
    form_view_extra_js = ['js/commons.js', 'js/form.js']
    form_view_extra_css = ['css/form.css']

    def index_view(self, request):
        kwargs = {'model_admin': self}
        view_class = self.dashboard_view_class

        request.session.setdefault('segment_view', 'dashboard')
        if request.session['segment_view'] != 'dashboard':
            view_class = self.index_view_class

        return view_class.as_view(**kwargs)(request)

    def page_count(self, obj):
        return len(obj.get_used_pages())

    def variant_count(self, obj):
        return len(obj.get_created_variants())

    def statistics(self, obj):
        return _("{visits} visits in {days} days").format(
            visits=obj.visit_count, days=obj.get_active_days())


def toggle_segment_view(request):
    """Toggle between the list view and dashboard view.

    :param request: The http request
    :type request: django.http.HttpRequest
    :returns: A redirect to the original page
    :rtype: django.http.HttpResponseRedirect

    """
    if request.user.has_perm('wagtailadmin.access_admin'):
        if request.session['segment_view'] == 'dashboard':
            request.session['segment_view'] = 'list'

        elif request.session['segment_view'] != 'dashboard':
            request.session['segment_view'] = 'dashboard'

        return HttpResponseRedirect(request.META.get('HTTP_REFERER', '/'))

    return HttpResponseForbidden()


def toggle(request, segment_id):
    """Toggle the status of the selected segment.

    :param request: The http request
    :type request: django.http.HttpRequest
    :param segment_id: The primary key of the segment
    :type segment_id: int
    :returns: A redirect to the original page
    :rtype: django.http.HttpResponseRedirect

    """
    if request.user.has_perm('wagtailadmin.access_admin'):
        segment = get_object_or_404(Segment, pk=segment_id)

        segment.toggle()

        return HttpResponseRedirect(request.META.get('HTTP_REFERER', '/'))

    return HttpResponseForbidden()


def copy_page_view(request, page_id, segment_id):
    """Copy page with selected segment. If the page for the selected segment
    already exists the user will be redirected to that particular page.

    :param request: The http request
    :type request: django.http.HttpRequest
    :param page_id: The primary key of the page
    :type segment_id: int
    :param segment_id: The primary key of the segment
    :type segment_id: int
    :returns: A redirect to the new page
    :rtype: django.http.HttpResponseRedirect

    """
    if request.user.has_perm('wagtailadmin.access_admin'):
        segment = get_object_or_404(Segment, pk=segment_id)
        page = get_object_or_404(Page, pk=page_id).specific

        metadata = page.personalisation_metadata
        variant_metadata = metadata.metadata_for_segments([segment])
        if variant_metadata.exists():
            variant = variant_metadata.first()
        else:
            variant = metadata.copy_for_segment(segment)
        edit_url = reverse('wagtailadmin_pages:edit', args=[variant.id])

        return HttpResponseRedirect(edit_url)

    return HttpResponseForbidden()


# CSV download views
def segment_user_data(request, segment_id):
    if request.user.has_perm('wagtailadmin.access_admin'):
        segment = get_object_or_404(Segment, pk=segment_id)

        response = HttpResponse(content_type='text/csv; charset=utf-8')
        response['Content-Disposition'] = \
            'attachment;filename=segment-%s-users.csv' % str(segment_id)

        headers = ['Username']
        for rule in segment.get_rules():
            if rule.static:
                headers.append(rule.get_column_header())

        writer = csv.writer(response)
        writer.writerow(headers)

        for user in segment.static_users.all():
            row = [user.username]
            for rule in segment.get_rules():
                if rule.static:
                    row.append(rule.get_user_info_string(user))
            writer.writerow(row)

        return response

    return HttpResponseForbidden()
